/**
 * @file
 * $Id: daemon.c,v 1.63 2003/05/13 12:44:43 tsavola Exp $
 * @author Timo Savola <timo.savola@movial.fi>
 */

#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include <syslog.h>
#include <pwd.h>
#include <netdb.h>
#include <signal.h>
#include <errno.h>
#include <pty.h>
#include <utmp.h>
#include <fcntl.h>
#include <getopt.h>
#include <time.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <netinet/in.h>
#include "daemon.h"
#include "common.h"
#include "protocol.h"
#include "buffer.h"

#ifndef MAX
#define MAX(a, b)                ((a) > (b) ? (a) : (b))
#endif

/** Calls debugfunc if debug_file is set. */
#define DEBUG                    if (debug_file) print_debug

#define send_empty(data, ptype)  write_packet((data)->sd, (ptype), NULL, 0)

/* The environment of the process. */
extern char **environ;

/** Debug is enabled if this is set. */
static FILE *debug_file = NULL;
static pthread_mutex_t debug_mutex = PTHREAD_MUTEX_INITIALIZER;

static char *mount_cmd = DEFAULT_MOUNT_CMD;
static char *umount_cmd = DEFAULT_UMOUNT_CMD;
static char *mount_tab = DEFAULT_MOUNT_TAB;
static char *bind_opt = DEFAULT_BIND_OPT;

/** The server socket. */
static int srvsd = -1;

/** Set of signals blocked by handler threads (SIGCHLD). */
static sigset_t sigs_to_block;

/** Used when calling the mount functions. */
static pthread_mutex_t mount_mutex = PTHREAD_MUTEX_INITIALIZER;

/** Doubly linked list of mount entries. */
static struct { mount_t *head, *tail; } mounts = { NULL, NULL };

/**
 * Prints only timestamp and pid to debug log.
 */
static void print_debug_prefix(void)
{
	struct timeval tv;
	struct tm *t;

	gettimeofday(&tv, NULL);
	t = gmtime(&tv.tv_sec);

	fprintf(debug_file,
		"%02d-%02d-%04d %02d:%02d:%02d.%03ld %05d ",
		t->tm_mday, t->tm_mon + 1, 1900 + t->tm_year,
		t->tm_hour, t->tm_min, t->tm_sec,
		tv.tv_usec / 1000,
		getpid());
}

/**
 * Prints message to debug log.
 */
static void print_debug(const char *msg, ...)
{
	va_list arg;

	pthread_mutex_lock(&debug_mutex);

	print_debug_prefix();

	va_start(arg, msg);
	vfprintf(debug_file, msg, arg);
	va_end(arg);

	fprintf(debug_file, "\n");
	fflush(debug_file);

	pthread_mutex_unlock(&debug_mutex);
}

/**
 * Prints command line to debug log.
 */
static void print_debug_cmdline(cmd_info_t *cmd)
{
	char **p;

	pthread_mutex_lock(&debug_mutex);

	print_debug_prefix();
	fprintf(debug_file, "Executing command: %s", cmd->cmd);

	p = cmd->argv;
	if (p) {
		while (*++p) {
			fprintf(debug_file, " %s", *p);
		}
	}

	fprintf(debug_file, "\n");
	fflush(debug_file);

	pthread_mutex_unlock(&debug_mutex);
}

/**
 * Prints message and errno description to syslog.
 */
static void error(const char *msg)
{
	char *desc = strerror(errno);

	if (!errno) {
		syslog(LOG_ERR, msg);
		DEBUG("Error: %s", msg);
		return;
	}

	syslog(LOG_ERR, "%s (%s)", msg, desc);
#ifdef NDEBUG
	DEBUG("Error: %s (%s)", msg, desc);
#else
	DEBUG("Error: %s (%s, %d)", msg, desc, errno);
#endif
}

/**
 * Prints message and errno description to syslog and stderr.
 */
static void error_err(const char *progname,
		      const char *msg)
{
	char *desc = strerror(errno);

	if (!errno) {
		fprintf(stderr, "%s: %s\n", progname, msg);
		DEBUG("Error: %s", msg);
		return;
	}

	fprintf(stderr, "%s: %s (%s)\n", progname, msg, desc);
#ifdef NDEBUG
	DEBUG("Error: %s (%s)", msg, desc);
#else
	DEBUG("Error: %s (%s, %d)", msg, desc, errno);
#endif
}

/**
 * Prints message to syslog.
 */
static void warn(const char *msg)
{
	syslog(LOG_WARNING, msg);
	DEBUG("Warning: %s", msg);
}

/**
 * Prints message and errno description to syslog and sends it to the client.
 * @param data if null, behaves like error()
 */
static void send_error(handler_t *data,
		       const char *msg)
{
	int size;
	char str[256];

	if (!data) {
		error(msg);
		return;
	}

	if (!errno) {
		error(msg);
		write_packet(data->sd, PTYPE_ERROR, (void *) msg, strlen(msg));
		return;
	}

	size = string_cat(str, sizeof (str),
			  msg, " (", strerror(errno), ")", NULL);

	syslog(LOG_ERR, str);
	DEBUG("Error: %s", str);
	write_packet(data->sd, PTYPE_ERROR, str, size);
}

/**
 * Sends message to the client.
 */
static void send_message(handler_t *data,
			 const char *msg)
{
	DEBUG("Sending message: %s", msg);
	write_packet(data->sd, PTYPE_MESSAGE, (char *) msg, strlen(msg));
}

/**
 * Send a SIGKILL the calling process. This function never returns.
 */
static void commit_suicide(void)
{
	pid_t pid;

	pid = getpid();
	if (pid >= 0) {
		kill(pid, SIGKILL);
	}

	/* If getpid or kill failed for some reason.. */
	abort();
}

/**
 * Passes the contents of BUF to send_error().
 */
static void flush(handler_t *data,
		  char *buf,
		  int *lenp)
{
	if (*lenp > 0) {
		buf[*lenp] = '\0';
		*lenp = 0;

		send_error(data, buf);
	}
}

/**
 * Executes argv[0] in a child process.
 * @return -1 on error, 0 otherwise
 */
static int execute(handler_t *data,
		   char **argv)
{
	int pid, status, err[2], pos = 0;
	char line[1024];

	if (pipe(err) < 0) {
		send_error(data, "Can't create pipe");
		return -1;
	}

	pid = fork();
	if (pid < 0) {
		return -1;
	}

	if (pid == 0) {
		/* Child */

		close(STDIN_FILENO);
		close(STDOUT_FILENO);
		if (dup2(err[1], STDERR_FILENO) != STDERR_FILENO) {
			send_error(data, "Can't duplicate pipe as stderr");
			commit_suicide();
		}
		close(err[0]);

		/* We want to write error to socket if exec fails. */
		set_closeonexec(data->sd);
		set_closeonexec(srvsd);

		execv(argv[0], argv);

		send_error(data, "Can't execute command");
		commit_suicide();
	}

	/* Parent */

	close(err[1]);

	while (1) {
		char c;

		if (read(err[0], &c, 1) <= 0) {
			flush(data, line, &pos);
			break;
		}

		if (c == '\n' || c == '\0') {
			flush(data, line, &pos);
			continue;
		}

		if (pos == sizeof (line)) {
			flush(data, line, &pos);
		}

		line[pos++] = c;
	}

	if (waitpid(pid, &status, 0) != pid) {
		return -1;
	}

	if (status < 0) {
		return -1;
	}

	if (WEXITSTATUS(status) != 0) {
		errno = 0;
		return -1;
	}

	return 0;
}

/**
 * Finds a host from a config file and returns the associated password.
 *
 * Example config file entry:
 * 192.168.0.2  foobar
 *
 * @param filename the config file path
 * @param host name of the remote machine
 * @return the password if found, NULL if not
 */
static char *find_pwd(handler_t *data,
		      const char *filename,
		      const char *host)
{
	FILE *file;
	char buf[1024], *buf2, *pwd = NULL;
	int len;

	file = fopen(filename, "r");
	if (!file) {
		send_error(data, "Can't open config file");
		return NULL;
	}

	while (1) {
		len = read_line(file, buf, sizeof (buf));
		if (len < 0) {
			break;
		}
		if (len == 0 || isspace(buf[0])) {
			continue;
		}

		buf2 = find_space(buf);
		if (*buf2 == '\0') {
			continue;
		}
		*buf2++ = '\0';

		if (strcmp(buf, host) != 0) {
			continue;
		}

		buf2 = skip_spaces(buf2);
		if (*buf2 == '\0') {
			continue;
		}
		*find_space(buf2) = '\0';

		pwd = strdup(buf2);
		if (!pwd) {
			send_error(data, "Out of memory");
		}
		break;
	}

	fclose(file);

	return pwd;
}

/**
 * Recursively creates all needed directories (with MKDIR_PERMS).
 */
static int mkdirs(const char *path)
{
	char *parent, *p;
	int rc = -1;

	if (mkdir(path, MKDIR_PERMS) == 0 || errno == EEXIST) {
		return 0;
	}

	if (errno != ENOENT) {
		return -1;
	}

	parent = strdup(path);
	if (!parent) {
		return -1;
	}

	p = strrchr(parent, '/');
	if (p && parent != p) {
		*p = '\0';
		rc = 0;
	}

	if (rc == 0) {
		rc = mkdirs(parent);
	}

	free(parent);

	if (rc == 0) {
		rc = mkdir(path, MKDIR_PERMS);
	}

	return rc;
}

/**
 * Frees all resources associated with a mount entry.
 */
static void mnt_free(mount_t *m)
{
	if (m->opts) {
		free(m->opts);
	}

	if (m->device) {
		free(m->device);
	}

	if (m->point) {
		free(m->point);
	}

	free(m);
}

/**
 * Deep-copies a mount entry.
 */
static mount_t *mnt_copy(mount_t *mnt)
{
	mount_t *c;

	c = calloc(sizeof (mount_t), 1);
	if (!c) {
		return NULL;
	}

	c->type = mnt->type;

	c->point = strdup(mnt->point);
	if (!c->point) {
		mnt_free(mnt);
		return NULL;
	}

	c->device = strdup(mnt->device);
	if (!c->device) {
		mnt_free(mnt);
		return NULL;
	}

	if (mnt->opts) {
		c->opts = strdup(mnt->opts);
		if (!c->opts) {
			mnt_free(mnt);
			return NULL;
		}
	}

	return c;
}

/**
 * Parses a string of form: <type> <device> <point> [<options>]
 */
static mount_t *mnt_parse(handler_t *data,
			  char *str,
			  const char *rootdir)
{
	char *typestr, *dev, *pnt, *opt;
	int type;
	mount_t *m;

	split_string(str, &typestr, &dev, &pnt, &opt, NULL);

	if (!typestr) {
		send_error(data, "No type in mount string");
		return NULL;
	}
	if (!dev) {
		send_error(data, "No device in mount string");
		return NULL;
	}
	if (!pnt) {
		send_error(data, "No mount point in mount string");
		return NULL;
	}

	if (strcmp(typestr, "nfs") == 0) {
		type = MTYPE_NFS;
	} else if (strcmp(typestr, "bind") == 0) {
		type = MTYPE_BIND;
	} else {
		send_error(data, "Invalid mount type");
		return NULL;
	}

	m = calloc(sizeof (mount_t), 1);
	if (!m) {
		send_error(data, "Out of memory");
		return NULL;
	}

	m->type = type;

	m->device = strdup(dev);
	if (!m->device) {
		send_error(data, "Out of memory");
		mnt_free(m);
		return NULL;
	}

	m->point = malloc(strlen(rootdir) + strlen(pnt) + 1);
	if (!m->point) {
		send_error(data, "Out of memory");
		mnt_free(m);
		return NULL;
	}
	strcpy(m->point, rootdir);
	strcat(m->point, pnt);

	/* Remove / from the end. */
	str = strrchr(m->point, '/');
	if (str && strlen(str) == 1) {
		*str = '\0';
	}

	if (opt) {
		m->opts = strdup(opt);
		if (!m->opts) {
			send_error(data, "Out of memory");
			mnt_free(m);
			return NULL;
		}
	} else {
		m->opts = NULL;
	}

	return m;
}

/**
 * Finds an entry from the MOUNTS list.
 * @param point the mount point of the entry
 */
static mount_t *mnt_list_find(const char *point)
{
	mount_t *mnt;

	for (mnt = mounts.head; mnt; mnt = mnt->next) {
		if (strcmp(mnt->point, point) == 0) {
			break;
		}
	}

	return mnt;
}

/**
 * Adds an entry to the MOUNTS list. The entries are kept in ascending order
 * based on their mount point strings.
 */
static void mnt_list_add(mount_t *mnt)
{
	mount_t *prev, *next;

	prev = NULL;
	next = mounts.head;
	while (next) {
		if (strcmp(next->point, mnt->point) > 0) {
			break;
		}

		prev = next;
		next = next->next;
	}

	mnt->prev = prev;
	mnt->next = next;

	if (prev) {
		prev->next = mnt;
	} else {
		mounts.head = mnt;
	}

	if (next) {
		next->prev = mnt;
	} else {
		mounts.tail = mnt;
	}
}

/**
 * Removes and frees an entry from the MOUNTS list.
 */
static void mnt_list_del(mount_t *mnt)
{
	if (mnt->prev) {
		mnt->prev->next = mnt->next;
	} else {
		mounts.head = mnt->next;
	}

	if (mnt->next) {
		mnt->next->prev = mnt->prev;
	} else {
		mounts.tail = mnt->prev;
	}

	mnt_free(mnt);
}

/**
 * Executes mount_cmd.
 */
static int mnt_mount(handler_t *data,
		     const mount_t *m)
{
	char *argv[8];

	memset(argv, 0, sizeof (argv));

	argv[0] = mount_cmd;
	argv[1] = m->device;
	argv[2] = m->point;

	switch (m->type) {
	case MTYPE_NFS:
		argv[3] = "-t";
		argv[4] = "nfs";
		argv[5] = m->opts ? "-o" : NULL;
		argv[6] = m->opts;

		if (m->opts) {
			DEBUG("Executing: %s %s %s -t nfs -o %s",
			      argv[0], argv[1], argv[2], argv[6]);
		} else {
			DEBUG("Executing: %s %s %s -t nfs",
			      argv[0], argv[1], argv[2]);
		}

		break;

	case MTYPE_BIND:
		split_string(bind_opt,
			     &argv[3], &argv[4], &argv[5], &argv[6], NULL);

		DEBUG("Executing: %s %s %s %s %s %s %s",
		      argv[0], argv[1], argv[2],
		      argv[3] ? argv[3] : "",
		      argv[4] ? argv[4] : "",
		      argv[5] ? argv[5] : "",
		      argv[6] ? argv[6] : "");

		break;

	default:
		errno = EINVAL;
		return -1;
	}

	return execute(data, argv);
}

static int unmount(handler_t *data,
		   char *point)
{
	char *argv[] = { umount_cmd, point, NULL };

	DEBUG("Executing: %s %s", argv[0], argv[1]);

	return execute(data, argv);
}

/**
 * Checks /etc/mtab if m->point is already mounted.
 * @param oldp will be set if already mounted
 * @return -1 on error, 0 if wasn't mounted, 1 if was mounted
 */
static int mnt_touch(handler_t *data,
		     mount_t *mnt,
		     mount_t **oldp)
{
	FILE *file;
	char buf[1024], *dev, *pnt, *type, *opt;
	int ok, len, mounted = 0;
	mount_t *old;

	/* Check the file. */

	file = fopen(mount_tab, "r");
	if (!file) {
		send_error(data, "Can't open mount table file");
		return -1;
	}

	while (1) {
		len = read_line(file, buf, sizeof (buf));
		if (len < 0) {
			break;
		}

		split_string(buf, &dev, &pnt, &type, &opt, NULL);
		if (!dev || !pnt) {
			continue;
		}

		/* Is this the right mount point? */
		if (strcmp(pnt, mnt->point) != 0) {
			continue;
		}

		/* Is device still the same? */
		ok = (strcmp(dev, mnt->device) == 0);

#if 0  /* mtab may show some implicit options, or someting ... */

		/* What about options? */
		if (ok) {
			if (opt && mnt->opts) {
				ok = (strcmp(opt, mnt->opts) == 0);
			} else {
				/* Both null? */
				ok = (opt == mnt->opts);
			}
		}

#endif

		if (ok) {
			mounted = 1;
		} else {
			mount_t *old;

			DEBUG("Unmounting old %s", pnt);
			unmount(data, pnt);

			old = mnt_list_find(pnt);
			if (old) {
				mnt_list_del(old);
			}
		}

		break;
	}

	fclose(file);

	/* Check the list. */

	old = mnt_list_find(mnt->point);

	if (mounted) {
		DEBUG("%s is mounted", mnt->point);

		if (old) {
			old->usage++;
		} else {
			old = mnt_copy(mnt);
			if (!old) {
				return -1;
			}

			old->usage = 1;
			mnt_list_add(old);
		}

		*oldp = old;
	} else if (old) {
		DEBUG("%s is not mounted but exists on our list", mnt->point);

		mnt_list_del(old);
	}

	return mounted;
}

/**
 * Mounts a filesystem if not already mounted.
 */
static mount_t *add_mount(handler_t *data,
			  char *str,
			  const char *rootdir)
{
	mount_t *mnt, *old = NULL;
	int mounted;

	mnt = mnt_parse(data, str, rootdir);
	if (!mnt) {
		return NULL;
	}

	mounted = mnt_touch(data, mnt, &old);
	if (mounted < 0) {
		mnt_free(mnt);
		return NULL;
	}

	if (mounted) {
		mnt_free(mnt);
		return old;
	}

	DEBUG("Creating directory %s with uid %d",
	      mnt->point, data->passwd.pw_uid);

	if (seteuid(data->passwd.pw_uid) < 0) {
		send_error(data, "Can't change uid");
		mnt_free(mnt);
		return NULL;
	}

	if (mkdirs(mnt->point) < 0) {
		send_error(data, "Can't create mount point");
		mnt_free(mnt);
		return NULL;
	}

	if (seteuid(0) < 0) {
		send_error(data, "Can't change uid back to root");
		mnt_free(mnt);
		return NULL;
	}

	if (mnt_mount(data, mnt) < 0) {
		mnt_free(mnt);
		return NULL;
	}

	mnt->usage = 1;
	mnt_list_add(mnt);

	return mnt;
}

/**
 * Unmounts a filesystem.
 * @return -1 on error, 0 otherwise
 */
static int remove_mount(handler_t *data,
			char *str,
			const char *rootdir)
{
	mount_t *mnt, *old = NULL;
	int mounted;

	mnt = mnt_parse(data, str, rootdir);
	if (!mnt) {
		return -1;
	}

	mounted = mnt_touch(data, mnt, &old);

	if (mounted >= 0 && old) {
		mnt_list_del(old);
	}

	if (mounted != 0) {
		unmount(data, mnt->point);
	}

	mnt_free(mnt);

	return (mounted < 0) ? -1 : 0;
}

/**
 * Decrements the usage count of a mount and possibly sets its expiration time.
 */
static void release_mount(mount_t *mnt)
{
	DEBUG("Releasing mount %s", mnt->point);

	mnt->usage--;

	if (mnt->usage <= 0) {
		mnt->expiration = time(NULL) + EXPIRATION;
	}
}

/**
 * Steps through the MOUNTS list and unmounts all unused and expired entries.
 */
static void expire_mounts(void)
{
	mount_t *prev, *curr;
	time_t now;

	now = time(NULL);

	prev = NULL;
	curr = mounts.tail;
	while (curr) {
		prev = curr->prev;

		if (curr->usage <= 0 &&
		    curr->expiration <= now) {
			DEBUG("Unmounting %s", curr->point);

			if (curr->next) {
				curr->next->prev = curr->prev;
			} else {
				mounts.tail = curr->prev;
			}

			if (curr->prev) {
				curr->prev->next = curr->next;
			} else {
				mounts.head = curr->next;
			}

			unmount(NULL, curr->point);
			mnt_free(curr);
		}

		curr = prev;
	}
}

/**
 * Unmounts all entries in the MOUNTS list.
 */
static void unmount_all(void)
{
	mount_t *prev, *mnt;

	mnt = mounts.tail;
	while (mnt) {
		prev = mnt->prev;

		DEBUG("Unmounting %s", mnt->point);

		unmount(NULL, mnt->point);
		mnt_free(mnt);

		mnt = prev;
	}

	mounts.head = NULL;
	mounts.tail = NULL;
}

/**
 * Gets the amount of data we can read from a file (which may be a tty).
 */
static int get_data_length(handler_t *data,
			   int fd,
			   int max)
{
	int len;

	len = max;

	if (isatty(fd)) {
		if (ioctl(fd, FIONREAD, &len) < 0) {
			send_error(data, "Can't check pty for available data");
			return -1;
		}

		if (len < 0) {
			send_error(data, "ioctl() gave invalid read length");
			return -1;
		}

		if (len > max) {
			len = max;
		}
	}

	return len;
}

/**
 * Reads some data from a fd and writes it into the socket (sd) in a packet.
 * The fd will be set to -1 at EOF.
 * @param fd pointer to the stdout or stderr of the child process
 * @param ptype the packet type
 * @return 0 if some data was copied or if the fd hit EOF,
 *         1 if no data was copied from a pty, -1 on error
 */
static int send_data(handler_t *data,
		     int *fd,
		     int ptype,
		     int *req)
{
	int len;

	len = get_data_length(data, *fd, *req);
	if (len < 0) {
		return -1;
	}
	if (len == 0) {
		return 1;
	}

	len = read(*fd, data->tmp_buf, len);
	if (len < 0) {
		send_error(data, "Can't read output of the process");
		return -1;
	}

	if (len) {
		DEBUG(ptype == PTYPE_OUT_DATA ?
		      "Sending OUT DATA packet" : "Sending ERR DATA packet");

		if (write_packet(data->sd, ptype, data->tmp_buf, len) < 0) {
			error("Can't write packet to socket");
			return -1;
		}

		(*req) -= len;
	} else {
		DEBUG(ptype == PTYPE_OUT_DATA ?
		      "Stdout hit EOF" : "Stderr hit EOF");

		*fd = -1;
		*req = 0;
	}

	return 0;
}

/**
 * Writes an RC or AUTH OK packet.
 */
static int send_int16(handler_t *data,
		      int ptype,
		      int16_t val)
{
	val = htons(val);

	if (write_packet(data->sd, ptype, &val, sizeof (val)) < 0) {
		error("Can't write int16 packet to socket");
		return -1;
	}

	return 0;
}

/**
 * Sends a request for more IN DATA.
 * @return 0 on success, -1 on error
 */
static int send_request(handler_t *data)
{
	DEBUG("Sending IN REQ packet");

	if (write_packet(data->sd, PTYPE_IN_REQ, NULL, 0) < 0) {
		error("Can't write packet to socket");
		return -1;
	}

	data->inwait = 1;

	return 0;
}

/**
 * Writes (some of) buf_in to a file.
 * @return 0 on success, -1 on error
 */
static int write_buffer(handler_t *data,
			int *fd)
{
	DEBUG("Writing buffer to stdin");

	if (buf_write_out(&data->buf_in, fd) < 0) {
		send_error(data, "Can't write buffer to stdin");
		return -1;
	}

	if (buf_is_empty(&data->buf_in)) {
		data->inwait = 0;
	}

	return 0;
}

/**
 * Reads bytes from socket to buf_in.
 * @param len the maximum number of bytes to be read
 * @return usually 0, 1 at EOF (len was zero), -1 on error
 */
static int receive_stream(handler_t *data,
			  int len)
{
	if (len) {
		DEBUG("Receiving IN DATA packet");

		if (buf_read_in(&data->buf_in, data->sd, len) < 0) {
			send_error(data,"Can't read IN DATA packet to buffer");
			return -1;
		}

		return 0;
	} else {
		DEBUG("Receiving IN DATA packet: EOF");

		buf_set_eof(&data->buf_in);

		return 1;
	}
}

/**
 * Read an IN DATA packet from socket (sd).
 * @param fd is the pointer to the file where the data is to be copied;
 *           it will be set to -1 if IN DATA packet was empty (EOF)
 * @return 0 on success, -1 on error
 */
static int receive_packet(handler_t *data,
			  int *fd)
{
	phead_t head;

	if (read_phead(data->sd, &head) < 0) {
		error("Can't read packet header from socket");
		return -1;
	}

	switch (head.type) {
	case PTYPE_IN_DATA:
		return receive_stream(data, head.size);

	case PTYPE_OUT_REQ:
		DEBUG("Receiving OUT REQ packet");

		if (head.size) {
			send_error(data, "OUT REQ packet size was non-zero");
			return -1;
		}

		data->outreq = BUFFER_SIZE;
		return 0;

	case PTYPE_ERR_REQ:
		DEBUG("Receiving ERR REQ packet");

		if (head.size) {
			send_error(data, "ERR REQ packet size was non-zero");
			return -1;
		}

		data->errreq = BUFFER_SIZE;
		return 0;
	}

	send_error(data, "Received packet has invalid type");
	return -1;
}

/**
 * Prints information on an exited process (pid) to debug log.
 */
static void print_status(int pid,
			 int status)
{
	if (WIFEXITED(status)) {
		DEBUG("PID %d exited normally: %d", pid, WEXITSTATUS(status));
	} else if (WIFSIGNALED(status)) {
		DEBUG("PID %d was signaled: %d", pid, WTERMSIG(status));
	} else if (WIFSTOPPED(status)) {
		DEBUG("PID %d was stopped: %d", pid, WSTOPSIG(status));
	} else {
		DEBUG("PID %d died due to unknown reason", pid);
	}
}

/**
 * Manages the streams of the command executor (= the child-child process).
 * @param pid of the child process, used for getting the exit status
 * @return -1 on error (the child should be killed), otherwise 0
 */
static int manage(handler_t *data,
		  pid_t pid,
		  int infd,
		  int outfd,
		  int errfd,
		  int *rcp)
{
	int maxfd1;
	fd_set readfds;
	fd_set writefds;
	int alive = 1;
	int status = 0;
	uint16_t rc = 0;

	/* Find out the maximum fd + 1 for select(). */
	maxfd1 = MAX(data->sd, infd);
	maxfd1 = MAX(maxfd1, outfd);
	maxfd1 = MAX(maxfd1, errfd);
	maxfd1++;

	if (!buf_init(&data->buf_in)) {
		send_error(data, "Can't initialize buffer");
		return -1;
	}

	/* Copy packets around until we get SIGCHLD. */
	while (alive || outfd >= 0 || errfd >= 0) {
		int copyout, copyerr, ok;

		copyout = outfd >= 0 && (data->outreq || !alive);
		copyerr = errfd >= 0 && (data->errreq || !alive);

		FD_ZERO(&readfds);
		FD_SET(data->sd, &readfds);
		if (copyout) {
			FD_SET(outfd, &readfds);
		}
		if (copyerr) {
			FD_SET(errfd, &readfds);
		}

		FD_ZERO(&writefds);
		if (infd >= 0 && !(buf_is_empty(&data->buf_in) &&
				   data->inwait)) {
			FD_SET(infd, &writefds);
		}

		if (select(maxfd1, &readfds, &writefds, NULL, NULL) <= 0) {
			error("Can't select");
			return -1;
		}

		/* Child died and pty doesn't have anything for us? */
		if (!alive && data->cmd.is_tty && FD_ISSET(outfd, &readfds) &&
		    get_data_length(data, outfd, 1) == 0) {
			break;
		}

		/* Read REQ or DATA packet from socket. */
		if (FD_ISSET(data->sd, &readfds) &&
		    receive_packet(data, &infd) < 0) {
			return -1;
		}

		/* Send data from stdout. */
		if (copyout && FD_ISSET(outfd, &readfds) &&
		    send_data(data,&outfd,PTYPE_OUT_DATA,&data->outreq) < 0) {
			return -1;
		}

		/* Send data from stderr. */
		if (copyerr && FD_ISSET(errfd, &readfds) &&
		    send_data(data,&errfd,PTYPE_ERR_DATA,&data->errreq) < 0) {
			return -1;
		}

		/* Flush buf_in to stdin or request more data. */
		if (infd >= 0 && FD_ISSET(infd, &writefds)) {
			if (buf_is_empty(&data->buf_in)) {
				ok = data->inwait || (send_request(data) >= 0);
			} else {
				ok = (write_buffer(data, &infd) >= 0);
			}

			if (!ok) {
				return -1;
			}
		}

		/* Did the child die already? */
		if (alive) {
			ok = waitpid(pid, &status, WNOHANG);
			if (ok < 0) {
				send_error(data, "Can't wait for child");
				return -1;
			}
			if (ok == pid) {
				alive = 0;
			}
		}
	}

	rc = 1;
	if (WIFEXITED(status)) {
		rc = WEXITSTATUS(status);
	}
	print_status(pid, status);

	*rcp = rc;
	return 0;
}

/**
 * @return the truth or -1 on error
 */
static int authenticate(handler_t *data)
{
	char *path, *pwd;
	int ok;

	path = malloc(strlen(data->passwd.pw_dir) +
		      strlen("/" CONFIG_NAME) + 1);
	if (!path) {
		send_error(data, "Out of memory");
		return -1;
	}
	strcpy(path, data->passwd.pw_dir);
	strcat(path, "/" CONFIG_NAME);

	DEBUG("Searching for password in %s", path);

	pwd = find_pwd(data, path, data->host);

	free(path);

	if (!pwd) {
		warn("Can't find password from config file");
		return 0;
	}

	ok = (strcmp(pwd, data->auth.pwd) == 0);

	free(pwd);

	if (!ok) {
		warn("Passwords don't match");
	}
	return ok;
}

/**
 * More or less like forkpty.
 * @return pid of the child process (0 for child) or -1 on error
 */
static int fork_pty(handler_t *data,
		    int *ptyfd)
{
	int master, slave, pid;
	struct winsize ws;

	ws.ws_row = data->cmd.tty_row;
	ws.ws_col = data->cmd.tty_col;
	ws.ws_xpixel = data->cmd.tty_xpixel;
	ws.ws_ypixel = data->cmd.tty_ypixel;

	if (openpty(&master, &slave, NULL, NULL, &ws) < 0) {
		send_error(data, "Can't open a pseudo-tty");
		return -1;
	}

	pid = fork();
	if (pid < 0) {
		send_error(data, "Can't fork");
		return -1;
	}

	if (pid == 0) {
		/* Child */

		close(master);
		if (login_tty(slave)) {
			send_error(data, "Can't login to a pseudo-tty");
			commit_suicide();
		}
	} else {
		/* Parent */

		close(slave);
		*ptyfd = master;
	}

	return pid;
}

/**
 * More or less like fork, but the stdin/stdout/stderr of the child process
 * are set to point to sockets.
 * @param infd a place to write the parent-end of the child's stdin socketpair
 * @return pid of the child process (0 for child) or -1 on error
 */
static int fork_sockets(handler_t *data,
			int *infd,
			int *outfd,
			int *errfd)
{
	int insd[2], outsd[2], errsd[2], pid;

	if (socketpair(AF_UNIX, SOCK_STREAM, 0, insd) < 0 ||
	    socketpair(AF_UNIX, SOCK_STREAM, 0, outsd) < 0 ||
	    socketpair(AF_UNIX, SOCK_STREAM, 0, errsd) < 0) {
		send_error(data, "Can't create socket pairs");
		return -1;
	}

	if (set_nonblocking(insd[0])) {
		send_error(data, "Can't make socket non-blocking");
		return -1;
	}

	pid = fork();
	if (pid < 0) {
		send_error(data, "Can't fork");
		return -1;
	}

	if (pid == 0) {
		/* Child */

		close(insd[0]);
		close(outsd[0]);
		close(errsd[0]);

		if (dup2(insd[1], STDIN_FILENO) != STDIN_FILENO) {
			send_error(data, "Can't duplicate socket as stdin");
			commit_suicide();
		}

		if (dup2(outsd[1], STDOUT_FILENO) != STDOUT_FILENO) {
			send_error(data, "Can't duplicate socket as stdout");
			commit_suicide();
		}

		if (dup2(errsd[1], STDERR_FILENO) != STDERR_FILENO) {
			send_error(data, "Can't duplicate socket as stderr");
			commit_suicide();
		}
	} else {
		/* Parent */

		close(insd[1]);
		close(outsd[1]);
		close(errsd[1]);

		*infd = insd[0];
		*outfd = outsd[0];
		*errfd = errsd[0];
	}

	return pid;
}

/**
 * Unmounts all filesystems listed in command info's option vector.
 * @return 0 or -1 on error
 */
static int unmount_opts(handler_t *data)
{
	char **optp;
	int cnt, error = 0;

	DEBUG("Unmounting filesystems");

	pthread_mutex_lock(&mount_mutex);

	/* Count the options and go to end of the vector. */
	for (cnt = 0, optp = data->cmd.opts; *optp; cnt++, optp++);

	/* Try to unmount filesystems in reverse order. */
	for (optp--; cnt-- > 0; optp--) {
		if (remove_mount(data, *optp, data->root) < 0) {
			error = -1;
		}
	}

	pthread_mutex_unlock(&mount_mutex);

	return error;
}

/**
 * Handles one client connection (in a child process).
 * @return -1 on error, 0 otherwise
 */
static int handle(handler_t *data)
{
	struct passwd *passwdp = NULL;
	int16_t ok;
	int count;
	char **optp;
	mount_t **mntp;

	DEBUG("Client IP is %s", data->host);

	DEBUG("Reading auth info struct from socket");

	if (read_auth(data->sd, &data->auth) < 0) {
		error("Can't read authentication info");
		return -1;
	}

	DEBUG("Client ID is %s", data->auth.client_id);

	DEBUG("Getting passwd struct of user %s", data->auth.user);

	if (getpwnam_r(data->auth.user, &data->passwd,
		       data->tmp_buf, BUFFER_SIZE, &passwdp) < 0) {
		send_error(data, "Can't get passwd structure");
		return -1;
	}

	ok = authenticate(data);
	if (ok < 0) {
		return -1;
	}

	if (ok) {
		DEBUG("Sending AUTH OK packet");

		if (send_empty(data, PTYPE_AUTH_OK) < 0) {
			error("Can't write AUTH OK packet to socket");
			return -1;
		}
	} else {
		send_message(data, "Permission denied");
		return -1;
	}

	DEBUG("Reading command info struct from socket");

	if (read_cmd(data->sd, &data->cmd) < 0) {
		error("Can't read command info");
		return -1;
	}

	/* Build path to new root. */

	data->root = malloc(strlen(data->passwd.pw_dir) + 1 +
			    strlen(data->host) + 1 + 
			    strlen(data->cmd.target) + 1);
	if (!data->root) {
		send_error(data, "Out of memory");
		return -1;
	}

	strcpy(data->root, data->passwd.pw_dir);
	strcat(data->root, "/");
	strcat(data->root, data->host);
	strcat(data->root, "-");
	strcat(data->root, data->cmd.target);

	/* Should we just unmount all filesystems and exit? */
	if (data->cmd.umount) {
		return unmount_opts(data);
	}

	/* Mount filesystems. */

	count = 0;
	for (optp = data->cmd.opts; *optp; optp++) {
		count++;
	}

	data->mounts = calloc(count + 1, sizeof (mount_t *));
	if (!data->mounts) {
		send_error(data, "Out of memory");
		return -1;
	}

	pthread_mutex_lock(&mount_mutex);

	mntp = data->mounts;
	for (optp = data->cmd.opts; *optp; optp++) {
		char errmsg[1024];
		mount_t *mnt;

		/* Just in case we need this later on. */
		string_cat(errmsg, sizeof (errmsg),
			   "Can't process mount entry: ", *optp, NULL);

		mnt = add_mount(data, *optp, data->root);
		if (!mnt) {
			pthread_mutex_unlock(&mount_mutex);

			errno = 0;
			send_error(data, errmsg);
			return -1;
		}

		*mntp++ = mnt;
	}

	pthread_mutex_unlock(&mount_mutex);

	/* Fork, etc. */

	if (data->cmd.is_tty) {
		/* TTY: */

		int fd, pid;

		DEBUG("Creating command process in a pty");

		pid = fork_pty(data, &fd);
		if (pid) {
			/* Parent: */

			int rc = -1;

			if (pid > 0) {
				if (manage(data, pid, fd, fd, -1, &rc) < 0) {
					if (kill(pid, SIGTERM) < 0) {
						error("Can't kill command");
					}
				}
			}

			close(fd);

			return rc;
		}
	} else {
		/* No TTY: */

		int infd, outfd, errfd, pid;

		DEBUG("Creating command process without a pty");

		pid = fork_sockets(data, &infd, &outfd, &errfd);
		if (pid) {
			/* Parent: */

			int rc = -1;

			if (pid > 0) {
				if (manage(data, pid,
					   infd, outfd, errfd, &rc) < 0) {
					if (kill(pid, SIGTERM) < 0) {
						error("Can't kill command");
					}
				}
			}

			close(errfd);
			close(outfd);
			close(infd);

			return rc;
		}
	}

	/* Child: */
	/* We'll kill the child on error so that cleanup() is not called. */

	signal(SIGCHLD, SIG_IGN);

	DEBUG("Changing root directory to %s", data->root);

	if (chroot(data->root) < 0) {
		send_error(data, "Can't change root directory");
		commit_suicide();
	}

	DEBUG("Changing current directory to %s", data->cmd.curdir);

	if (chdir("/") < 0 || chdir(data->cmd.curdir) < 0) {
		send_error(data, "Can't change current directory");
		commit_suicide();
	}

	DEBUG("Changing uid to %d and gid to %d",
	      data->passwd.pw_uid, data->passwd.pw_gid);

	/* Do this while we're still root. */
	if (setgid(data->passwd.pw_gid) < 0 &&
	    setegid(data->passwd.pw_gid) < 0) {
		send_error(data, "Can't change group id");
		commit_suicide();
	}

	if (setuid(data->passwd.pw_uid) < 0 &&
	    seteuid(data->passwd.pw_uid) < 0) {
		send_error(data, "Can't change user id");
		commit_suicide();
	}

	/* Print command line to debug log. */
	if (debug_file) {
		print_debug_cmdline(&data->cmd);
	}

	/* Change environment. */
	environ = data->cmd.env;

	/* We want to be able to send errors if exec fails. */
	set_closeonexec(data->sd);
	set_closeonexec(srvsd);

	execvp(data->cmd.cmd, data->cmd.argv);

	send_error(data, "Can't execute command");
	commit_suicide();

 	/* For the pedantic compiler.. */
	return 0;
}

/**
 * Frees all resources used by a handler thread.
 */
static void free_handler(handler_t *data)
{
	buf_free(&data->buf_in);

	if (data->mounts) {
		mount_t **p;

		pthread_mutex_lock(&mount_mutex);

		for (p = data->mounts; *p; p++) {
			release_mount(*p);
		}

		pthread_mutex_unlock(&mount_mutex);

		free(data->mounts);
	}

	if (data->root) {
		free(data->root);
	}

	free_cmd(&data->cmd);
	
	close(data->sd);

	free(data);
}

/**
 * Entry point of the handler thread.
 */
static void handler_main(handler_t *data)
{
	int rc;

	pthread_sigmask(SIG_BLOCK, &sigs_to_block, NULL);

	rc = handle(data);
	if (rc < 0) {
		rc = 1;
	}

	DEBUG("Sending RC packet: %d", rc);

	/* Send the return code to the client. */
	if (send_int16(data, PTYPE_RC, rc) >= 0) {

		/* Ok, then do something until the client goes away. */
		if (data->sd >= 0) {
			fd_set fds;

			FD_ZERO(&fds);
			FD_SET(data->sd, &fds);

			while (select(data->sd + 1, &fds,NULL,NULL,NULL) > 0 &&
			       read(data->sd, data->tmp_buf, BUFFER_SIZE) > 0);
		}
	}

	free_handler(data);

	DEBUG("Handler thread exiting");
}

/**
 * Invoked at SIGCHLD signal. wait()s for zombie children.
 */
static void catch_sigchld(int sig)
{
	int pid, status;

	DEBUG("SIGCHLD caught");

	pid = wait(&status);
	if (pid > 0) {
		print_status(pid, status);
	}
}

static void catch_sigpipe(int sig)
{
	DEBUG("SIGPIPE caught");
}

/**
 * Called when daemon exits. Unmounts all filesystems that we've mounted.
 */
static void cleanup(void)
{
	signal(SIGCHLD, SIG_IGN);

	pthread_mutex_lock(&mount_mutex);
	unmount_all();
	pthread_mutex_unlock(&mount_mutex);

	DEBUG("Daemon exiting");
}

/**
 * Gets the port number and debug log name. Prints usage and exits on error.
 * @return port number
 */
static int read_args(char *progname,
		     int argc,
		     char **argv)
{
	struct option longopts[] = {
		{ "help",       no_argument,       0, 'h' },
		{ "version",    no_argument,       0, 'v' },
		{ "port",       required_argument, 0, 'p' },
		{ "debug",      required_argument, 0, 'd' },
		{ "mount-bin",  required_argument, 0, 'm' },
		{ "umount-bin", required_argument, 0, 'u' },
		{ "mount-tab",  required_argument, 0, 't' },
		{ "bind-opt",   required_argument, 0, 'b' },
		{ 0 }
	};
	int port = DEFAULT_PORT;
	char *debugname = NULL;

	while (1) {
		int c;

		c = getopt_long(argc, argv, "hvp:d:m:u:t:b:", longopts, NULL);
		if (c < 0) {
			break;
		}

		switch (c) {
		case 'p':
			port = atoi(optarg);
			break;

		case 'd':
			debugname = optarg;
			break;

		case 'm':
			mount_cmd = optarg;
			break;

		case 'u':
			umount_cmd = optarg;
			break;

		case 't':
			mount_tab = optarg;
			break;

		case 'b':
			bind_opt = optarg;
			break;

		case 'v':
			fprintf(stderr, DAEMON_VERSION "\n");
			exit(0);

		case 'h':
		case '?':
		default:
			fprintf(stderr, "Usage: %s"
				" [-p|--port <port>]"
				" [-d|--debug <path>]"
				" [-m|--mount-bin <path>]"
				" [-u|--umount-bin <path>]"
				" [-t|--mount-tab <path>]"
				" [-b|--bind-opt <options>]"
				"\n       %s"
				" -v|--version"
				"\n       %s"
				" -h|--help"
				"\n", progname, progname, progname);
			exit(0);
		}
	}

	if (port <= 0 || port >= 65536) {
		error_err(progname, "Invalid port number");
		exit(1);
	}

	if (debugname) {
		debug_file = fopen(debugname, "a");
		if (!debug_file) {
			error_err(progname, "Can't open debug log file");
		}
	}

	return port;
}

/**
 * Parse args, open logs, create a socket, accept connections, fork handlers.
 */
int main(int argc,
	 char **argv)
{
	char *progname;
	int port, clisd, yes = 1;
	struct sockaddr_in srvaddr;
	pid_t pid;
	struct timeval timeout;
	pthread_attr_t thread_attr;

	progname = get_progname(argv[0]);
	openlog(progname, LOG_PID, LOG_DAEMON);

	port = read_args(progname, argc, argv);

	/* Create a socket, bind it and listen to it. */
	srvsd = socket(PF_INET, SOCK_STREAM, 0);
	if (srvsd < 0) {
		error_err(progname, "Can't create socket");
		return 1;
	}

	if (setsockopt(srvsd, SOL_SOCKET, SO_REUSEADDR,
		       &yes, sizeof (yes)) < 0) {
		error_err(progname, "Can't set socket option (SO_REUSEADDR)");
		return 1;
	}

	memset((char *) &srvaddr, 0, sizeof (srvaddr));
	srvaddr.sin_family = AF_INET;
	srvaddr.sin_addr.s_addr = htonl(INADDR_ANY);
	srvaddr.sin_port = htons(port);

	if (bind(srvsd, (struct sockaddr *) &srvaddr, sizeof (srvaddr)) < 0) {
		error_err(progname, "Can't bind socket");
		return 1;
	}

	if (listen(srvsd, BACKLOG) < 0) {
		error_err(progname, "Can't listen with socket");
		return 1;
	}

#if 1
	/* Daemon mode; kill the parent, etc. */
	pid = fork();
	if (pid < 0) {
		error_err(progname, "Can't fork");
		return 1;
	}
	if (pid > 0) {
		printf("%d\n", pid);
		return 0;
	}

	setsid();
	chdir("/");
	umask(0);
#endif

	pthread_attr_init(&thread_attr);
	pthread_attr_setdetachstate(&thread_attr, PTHREAD_CREATE_DETACHED);

	atexit(cleanup);
	signal(SIGINT, exit);
	signal(SIGTERM, exit);
	signal(SIGHUP, exit);
	signal(SIGCHLD, catch_sigchld);
	signal(SIGPIPE, catch_sigpipe);

	sigemptyset(&sigs_to_block);
	sigaddset(&sigs_to_block, SIGCHLD);
	sigaddset(&sigs_to_block, SIGPIPE);

	timeout.tv_sec = TIMEOUT;
	timeout.tv_usec = 0;

	/* Accept connections from clients. */
	while (1) {
		fd_set fds;

		FD_ZERO(&fds);
		FD_SET(srvsd, &fds);

		DEBUG("Waiting for connection");

		if (select(srvsd + 1, &fds, NULL, NULL, &timeout) < 0) {
			error("Can't select");
			return 1;
		}

		/* Did we get a new connection? */
		if (FD_ISSET(srvsd, &fds)) {
			struct sockaddr_in addr;
			socklen_t len;
			pthread_t thread;
			handler_t *data;

			len = sizeof (addr);
			clisd = accept(srvsd, (struct sockaddr *) &addr, &len);
			if (clisd < 0) {
				error("Can't accept");
				return 1;
			}

			DEBUG("New connection");

			data = calloc(sizeof (handler_t), 1);
			if (!data) {
				error("Out of memory");
				close(clisd);
				continue;
			}

			/* This must be set before calling free_handler(). */
			data->sd = clisd;

			/* Get the IP address. */
			if (getnameinfo((struct sockaddr *) &addr, len,
					data->host, sizeof (data->host),
					NULL, 0, NI_NUMERICHOST)) {
				error("Can't get client's IP address");
				free_handler(data);
				continue;
			}

			if (pthread_create(&thread, &thread_attr,
					   (void *) handler_main,
					   (void *) data) < 0) {
				error("Can't create thread");
				free_handler(data);
			}
		}

		/* Is it time to expire mounts? */
		if (timeout.tv_sec == 0 && timeout.tv_usec == 0) {
			timeout.tv_sec = TIMEOUT;
			timeout.tv_usec = 0;

			pthread_mutex_lock(&mount_mutex);

			DEBUG("Checking for expired mounts");
			expire_mounts();

			pthread_mutex_unlock(&mount_mutex);
		}
	}

	return 0;
}
