/**
 * @file
 * $Id: protocol.c,v 1.17 2003/05/07 08:36:54 tsavola Exp $
 * @author Timo Savola <timo.savola@movial.fi>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <netinet/in.h>
#include "protocol.h"

/**
 * Writes an integer of type int16_t into a file (in network byte order).
 * The integer should be non-negative.
 * @param fd the file descriptor
 * @param i the integer
 * @return 0 on success, -1 on error
 */
int write_int16(int fd,
		int16_t i)
{
	int16_t data = htons(i);

	if (write_buf(fd, &data, sizeof (data)) < 0) {
		return -1;
	}

	return 0;
}

int write_int32(int fd,
		int32_t i)
{
	int32_t data = htonl(i);

	if (write_buf(fd, &data, sizeof (data)) < 0) {
		return -1;
	}

	return 0;
}

/**
 * Writes a string into a file.
 * @param fd the file descriptor
 * @param str the null-terminated string
 * @return 0 on success, -1 on error
 */
int write_str(int fd,
	      const char *str)
{
	int16_t len;

	if (!str) {
		static const char *empty = "";
		str = empty;
	}

	len = strlen(str);

	/* Write the length. */
	if (write_int16(fd, len) < 0) {
		return -1;
	}

	/* Write the string. */
	if (write_buf(fd, (void *) str, len) < 0) {
		return -1;
	}

	return 0;
}

/**
 * Writes a string vector into a file.
 * @param fd the file descriptor
 * @param strv the string vector
 * @return 0 on success, -1 on error
 */
int write_strv(int fd,
	       const char **strv)
{
	int strc = 0;
	const char **s;

	if (!strv) {
		static const char *empty[] = { NULL };
		strv = empty;
	}

	/* Calculate the length. */
	for (s = strv; *s; s++) {
		strc++;
	}

	/* Write the length. */
	if (write_int16(fd, strc) < 0) {
		return -1;
	}

	/* Write the strings. */
	for (s = strv; *s; s++) {
		if (write_str(fd, *s) < 0) {
			return -1;
		}
	}

	return 0;
}

int write_buf(int fd,
	      void *buf,
	      int len)
{
	int i = 0;

	if (len == 0) {
		return 0;
	}

	while (i < len) {
		int cnt;

		cnt = write(fd, buf + i, len - i);
		if (cnt < 0) {
			if (errno == EAGAIN) {
				continue;
			}
			return -1;
		}

		i += cnt;
	}

	return 0;
}

/**
 * Reads a complete buffer from a file.
 * @param fd a file descriptor
 * @param buf pointer to a buffer large enough
 * @param len the length of the buffer
 * @return 0 on success, -1 on i/o error
 */
int read_buf(int fd,
	     void *buf,
	     int len)
{
	int i = 0;

	if (len == 0) {
		return 0;
	}

	while (i < len) {
		int cnt;

		cnt = read(fd, buf + i, len - i);
		if (cnt < 0) {
			if (errno == EAGAIN) {
				continue;
			}
			return -1;
		}

		if (cnt == 0) {
			errno = 0;
			return -1;
		}

		i += cnt;
	}

	return 0;
}

/**
 * Reads an integer of type int16_t from a file (where it is assumed to be
 * in network byte order).
 * @param fd a file descriptor
 * @return -1 on i/o error (thus negative values should not be used)
 */
int16_t read_int16(int fd)
{
	int16_t data;

	if (read_buf(fd, &data, sizeof (data)) < 0) {
		return -1;
	}

	return ntohs(data);
}

int32_t read_int32(int fd)
{
	int32_t data;

	if (read_buf(fd, &data, sizeof (data)) < 0) {
		return -1;
	}

	return ntohl(data);
}

/**
 * Reads a string from a file. The returned string should be free()'d.
 * @param fd a file descriptor
 * @return NULL on error
 */
char *read_str(int fd)
{
	int16_t len;
	char *str;

	/* Read the length. */
	len = read_int16(fd);
	if (len < 0) {
		return NULL;
	}

	str = calloc(len + 1, 1);
	if (!str) {
		return NULL;
	}

	/* Read the string. */
	if (read_buf(fd, str, len) < 0) {
		free(str);
		return NULL;
	}

	return str;
}

/**
 * Reads a string vector from a file. The returned strings and the vector
 * should be set free(). The vector is null-terminated.
 * @param fd a file descriptor
 * @return NULL on error
 */
char **read_strv(int fd)
{
	int16_t strc;
	char **strv;
	int i;

	/* Read the length. */
	strc = read_int16(fd);
	if (strc < 0) {
		return NULL;
	}

	strv = calloc(strc + 1, sizeof (char *));
	if (!strv) {
		return NULL;
	}

	/* Read the strings. */
	for (i = 0; i < strc; i++) {
		strv[i] = read_str(fd);
		if (!strv[i]) {
			free(strv);
			return NULL;
		}
	}

	return strv;
}

/**
 * Writes an authentication info structure into a file.
 * @param fd a file descriptor
 * @param user the username
 * @param pwd the password
 * @return 0 on success, -1 on error
 */
int write_auth(int fd,
	       const char *client_id,
	       const char *user,
	       const char *pwd)
{
	auth_info_t auth;

	if (strlen(client_id) >= CLIENT_ID_SIZE) {
		fprintf(stderr, "Warning: client ID is too long; "
			"truncating to %d characters\n", CLIENT_ID_SIZE - 1);
	}
	if (strlen(user) >= USER_SIZE) {
		fprintf(stderr, "Warning: username is too long; "
			"truncating to %d characters\n", USER_SIZE - 1);
	}
	if (strlen(pwd) >= PWD_SIZE) {
		fprintf(stderr, "Warning: password is too long; "
			"truncating to %d characters\n", PWD_SIZE - 1);
	}

	memset(auth.client_id, '\0', CLIENT_ID_SIZE);
	strncpy(auth.client_id, client_id, CLIENT_ID_SIZE - 1);

	memset(auth.user, '\0', USER_SIZE);
	strncpy(auth.user, user, USER_SIZE - 1);

	memset(auth.pwd, '\0', PWD_SIZE);
	strncpy(auth.pwd, pwd, PWD_SIZE - 1);

	if (write_buf(fd, &auth, sizeof (auth)) < 0) {
		return -1;
	}

	return 0;
}

/**
 * Reads an authentication info structure from a file.
 * @param fd a file descriptor
 * @param auth pointer to some memory...
 * @return -1 on error
 */
int read_auth(int fd,
	      auth_info_t *auth)
{
	return read_buf(fd, auth, sizeof (auth_info_t));
}

/**
 * Writes a command info structure into a file.
 */
int write_cmd(int fd,
	      int16_t umount,
	      const char *target,
	      const char **opts,
	      const char *cmd,
	      const char **argv,
	      const char *curdir,
	      const char **env,
	      int16_t is_tty,
	      const struct winsize *ws)
{
	if (write_int16(fd, umount       ) < 0 ||
	    write_str  (fd, target       ) < 0 ||
	    write_strv (fd, opts         ) < 0 ||
	    write_str  (fd, cmd          ) < 0 ||
	    write_strv (fd, argv         ) < 0 ||
	    write_str  (fd, curdir       ) < 0 ||
	    write_strv (fd, env          ) < 0 ||
	    write_int16(fd, is_tty       ) < 0 ||
	    write_int16(fd, ws->ws_row   ) < 0 ||
	    write_int16(fd, ws->ws_col   ) < 0 ||
	    write_int16(fd, ws->ws_xpixel) < 0 ||
	    write_int16(fd, ws->ws_ypixel) < 0) {
		return -1;
	}

	return 0;
}

/**
 * Reads a command info structure from a file. The CMD structure
 * should be freed with the free_cmd() function.
 * @param fd a file descriptor
 * @param cmd pointer to some memory...
 * @return -1 on error
 */
int read_cmd(int fd,
	     cmd_info_t *cmd)
{
	cmd->umount = read_int16(fd);
	if (cmd->umount < 0) {
		free_cmd(cmd);
		return -1;
	}

	cmd->target = read_str(fd);
	if (!cmd->target) {
		free_cmd(cmd);
		return -1;
	}

	cmd->opts = read_strv(fd);
	if (!cmd->opts) {
		free_cmd(cmd);
		return -1;
	}

	cmd->cmd = read_str(fd);
	if (!cmd->cmd) {
		free_cmd(cmd);
		return -1;
	}

	cmd->argv = read_strv(fd);
	if (!cmd->argv) {
		free_cmd(cmd);
		return -1;
	}

	cmd->curdir = read_str(fd);
	if (!cmd->curdir) {
		free_cmd(cmd);
		return -1;
	}

	cmd->env = read_strv(fd);
	if (!cmd->env) {
		free_cmd(cmd);
		return -1;
	}

	cmd->is_tty = read_int16(fd);
	if (cmd->is_tty < 0) {
		free_cmd(cmd);
		return -1;
	}

	cmd->tty_row = read_int16(fd);
	if (cmd->tty_row < 0) {
		free_cmd(cmd);
		return -1;
	}

	cmd->tty_col = read_int16(fd);
	if (cmd->tty_col < 0) {
		free_cmd(cmd);
		return -1;
	}

	cmd->tty_xpixel = read_int16(fd);
	if (cmd->tty_xpixel < 0) {
		free_cmd(cmd);
		return -1;
	}

	cmd->tty_ypixel = read_int16(fd);
	if (cmd->tty_ypixel < 0) {
		free_cmd(cmd);
		return -1;
	}

	return 0;
}

/**
 * Frees a command info structure including its contents.
 */
void free_cmd(cmd_info_t *cmd)
{
	if (cmd->target) {
		free(cmd->target);
	}

	if (cmd->opts) {
		free_vec((void **) cmd->opts);
	}

	if (cmd->cmd) {
		free(cmd->cmd);
	}

	if (cmd->argv) {
		free_vec((void **) cmd->argv);
	}

	if (cmd->curdir) {
		free(cmd->curdir);
	}

	if (cmd->env) {
		free_vec((void **) cmd->env);
	}
}

/**
 * Writes a packet into a file.
 * @param fd a file descriptor
 * @param type of the packet (one of the PTYPE-values)
 * @param buf the contents of the packet
 * @param size the size of buf
 * @return 0 on success, -1 on i/o error
 */
int write_packet(int fd,
		 int16_t type,
		 void *buf,
		 int32_t size)
{
	if (write_int16(fd, type) < 0 ||
	    write_int32(fd, size) < 0) {
		return -1;
	}

	if (size > 0 && write_buf(fd, buf, size) < 0) {
		return -1;
	}

	return 0;
}

/**
 * Reads the header of a packet from a file so that it can be identified.
 * @param fd a file descriptor
 * @param head buffer for the packet
 * @return 0 on success, -1 on i/o error
 */
int read_phead(int fd,
	       phead_t *head)
{
	head->type = read_int16(fd);
	if (head->type < 0) {
		return -1;
	}

	head->size = read_int32(fd);
	if (head->size < 0) {
		return -1;
	}

	return 0;
}
